/**
  Weird assorted helpful include files for debugging and 
  testing assembly code.
  
  Orion Sky Lawlor, olawlor@acm.org, 2005/09/14 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>
#include <algorithm> /* for std::sort */
#include "inc.h"

/* To avoid cluttering up the screen during timing tests... */
int timer_only_dont_print=0;

/******* Function Performance Profiling ***********/
/**
  Return the current time in seconds (since something or other).
*/
double time_in_seconds_granularity=0.001; /* measure for 1 millisecond */
double time_in_seconds(void) { return MPI_Wtime(); }


CDECL int timeable_fn_empty(void) {
	return 0;
}

/**
  Return the number of seconds this function takes to run.
  May run the function several times (to average out 
  timer granularity).
*/
double time_function_onepass(timeable_fn fn)
{
	unsigned int i,count=1;
	double timePer=0;
	for (count=1;count!=0;count*=2) {
		double start, end, elapsed;
		timer_only_dont_print=1;
		start=time_in_seconds();
		for (i=0;i<count;i++) fn();
		if (doCudaSync) cudaThreadSynchronize();
		if (doGLfinish) glFinish();
		end=time_in_seconds();
		timer_only_dont_print=0;
		elapsed=end-start;
		MPI_Allreduce(&elapsed,&elapsed,1,MPI_DOUBLE, MPI_MAX,MPI_COMM_WORLD);
		timePer=elapsed/count;
		if (elapsed>time_in_seconds_granularity) /* Took long enough */
			return timePer;
	}
	/* woa-- if we got here, "count" reached integer wraparound before 
	  the timer ran long enough.  Return the last known time */
	return timePer;
}

/**
  Return the number of seconds this function takes to run.
  May run the function several times (to average out 
  timer granularity).
*/
double time_function(timeable_fn fn)
{
	static double empty_time=-1;
	if (empty_time<0) { /* Estimate overhead of subroutine call alone */
		empty_time=0; /* To avoid infinite recursion! */
		/* empty_time=time_function(timeable_fn_empty); */
	}
	enum {
#if defined(WIN32) /* Win32 timer has coarse granularity--too slow otherwise! */
		ntimes=1
#else
		ntimes=5
#endif
	};
	double times[ntimes];
	for (int t=0;t<ntimes;t++)
		times[t]=time_function_onepass(fn);
	std::sort(&times[0],&times[ntimes]);
	return times[ntimes/2]-empty_time;
}

/**
  Time a function's execution, and print this time out.
*/
void print_time(const char *fnName,timeable_fn fn)
{
	double sec=time_function(fn);
	printf("%s: ",fnName);
	if (1 || sec<1.0e-6) printf("%.2f ns/call\n",sec*1.0e9);
	else if (sec<1.0e-3) printf("%.2f us/call\n",sec*1.0e6);
	else if (sec<1.0e0) printf("%.2f ms/call\n",sec*1.0e3);
	else printf("%.2f s/call\n",sec);
}

